/*  A suite of programs for the analysis of chants.
   Copyright (C) 2000-2001 Ruth M. Ripley
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <windows.h>
#include <stdio.h>
#include <math.h>
#define max9(a,b) a>b?a:b
static double **matrix();
static int **imatrix();
static void free_matrix();
static void free_imatrix();
static int display=1;
		
static int errmsg(char *string, int fatal)
{
    int myresp;
    
    if (display || fatal)
    {
	myresp=MessageBox(0,string,"testbox",MB_OKCANCEL|MB_TASKMODAL);
    }
    
    if (myresp==IDCANCEL) 
	display=0;
    if (fatal)
	exit(0);
    return(myresp);
}

STDCALL void compS(int  *seq1, int *seq2, int len1, int len2, double *ss1,
		   double *S11, int *S10, double *sx, int sss,
		   int init, int metric, int *jm, int *im, double *maxs)
{
    int     i,j,allposs,l,**S0, imax, jmax;
    double sxy,**S,tmp,**ss, maxsim;
    double poss[3];
    char s[100];

    S = matrix(len1,len2);
    
    S0= imatrix(len1,len2);
    ss = matrix(sss+1,sss+1);
    for (j=1;j<=sss;j++)
	for (i=1;i<=sss;i++)
	    ss[i][j] = *ss1++;	   
    /* initialise*/
    if (init==1)
      {
	for(i=1;i<=len1;i++)
	  {
	    if (len1 < len2)
	      S[i][0]= S[i-1][0] - sx[seq1[i-1]-1];
/*	    S0[i][0] = 1;*/
	  }
	for(j=1;j<=len2;j++)
	  {
	    if (len2 < len1)
	      S[0][j]= S[0][j-1] - sx[seq2[j-1]-1];
	    /* S0[0][j] = 4; */
	  }
      }
    
    /* fill arrays*/
    maxsim=0.0;
    jmax=0;
    imax=0;
    if (metric==1)
      for (i=0;i<len1;i++)
	for (j=0;j<len2;j++)
	  {
	    sxy = ss[seq1[i]][seq2[j]];
	    poss[0] = S[i][j+1] - sx[seq1[i]-1];
	    poss[1] = S[i][j] + sxy;
	    poss[2] = S[i+1][j] - sx[seq2[j]-1];
	    
	    tmp = poss[0];
	    for (l=1;l<3;l++)
	      if (poss[l] > tmp)
		tmp = poss[l];
	    /*sprintf(s,"%d %d %f %f %f %f",i,j,poss[1],sxy,S[i][j],tmp);
	      errmsg(s);*/
	    allposs = 0;
	    for (l=0;l<3;l++)
	      if (fabs(poss[l] - tmp)<0.0001 )
		allposs+= pow(2.0,(double)l);
	    if ( metric==1 &&(fabs(tmp) < 0.0001 || tmp< 0.0))
	      {
		S[i+1][j+1] = 0;
		S0[i+1][j+1]= 0;
	      }
	    else
	      {
		S[i+1][j+1] = tmp;
		S0[i+1][j+1] = allposs;
		if (tmp > maxsim)
		  {
		    imax=i+1;
		    jmax=j+1;
		    maxsim=tmp;
		  }
	      }
	  }
    
    /*  sprintf(s," %d %d  %f %f %f %f %d %f %d %d %f %f\n",i,j,S[i+1][j+1],
	poss[0],poss[1],poss[2],allposs,sxy,seq1[i],seq2[j],s1[i],s2[j]);
	errmsg(s);	   */
    else
    {	
	for (i=0;i<len1;i++)
	    for (j=0;j<len2;j++)
	    {
		sxy = ss[seq1[i]][seq2[j]];
		poss[0] = S[i][j+1] - sx[seq1[i]-1];
		poss[1] = S[i][j] + sxy;
		poss[2] = S[i+1][j] - sx[seq2[j]-1];
		
		/*sprintf(s,"%d %d %f %f %f",i,j,poss[1],sxy,S[i][j]);
		  errmsg(s);*/
		
		tmp = poss[0];
		for (l=1;l<3;l++)
		    if (poss[l] > tmp)
			tmp = poss[l];
		/*sprintf(s,"%d %d %f %f %f %f",i,j,poss[1],sxy,S[i][j],tmp);
		  errmsg(s);*/
		allposs = 0;
		for (l=0;l<3;l++)
		    if (fabs(poss[l] - tmp)<0.000001 )
			allposs+= pow(2.0,(double)l);
		S[i+1][j+1] = tmp;
		S0[i+1][j+1] = allposs;
	    }
	if (metric==0)
	  {
	      imax=len1;
	      jmax=len2;
	      maxsim=tmp;
	  }
	else
	{	    
	    if (len2<len1)
	    {
		maxsim=S[len1][len2];
		imax=len1;
		jmax=len2;
		for (i=0;i<len1;i++)
		    if (maxsim < S[i][len2])
		    {
			imax=i;
			maxsim=S[i][len2];
		    }  
	    }
	    else
	    {
		maxsim=S[len1][len2];
		jmax=len2;
		imax=len1;
		for (j=0;j<len2;j++)
		    if (maxsim < S[len1][j])
		    {
			jmax=j;
			maxsim=S[len1][j];
		    }
	    }
	}
	
    }
    /*  sprintf(s," %d %d  %f %f %f %f %d %f %d %d %f %f\n",i,j,S[i+1][j+1],
	poss[0],poss[1],poss[2],allposs,sxy,seq1[i],seq2[j],s1[i],s2[j]);
	errmsg(s);	   */    
    /* copy S and S0 into the right places */
    for (j=0;j<=len2; j++)
      for (i=0;i<=len1;i++)
	{
	  *S10++ = S0[i][j];
	  *S11++ = S[i][j];
	}
    *jm=jmax;
    *im=imax;
    *maxs=maxsim;
    /*tmp = S[n][m];*/
    /*sprintf(s,"%f",4.2);*/
    /* errmsg(s);*/
     
    /* errmsg("end");      */    
    free_matrix(S,len1,len2);
    free_imatrix(S0,len1,len2);
    free_matrix(ss,sss+1,sss+1);
}

STDCALL void unravel(int ii, int jj, int *mlen, int *mtc, int *musd, 
		     int *eni, int *enj, 
		     int s1, int s2, int *S10, int len1, int len2)
     /*unravels from ii and jj, stores in mlen and matc arrays */
{
  int i, j, k, kk, iiuse, reli, relj, endi, endj, sets[4], **SST;
  int **matc, **mused, mn, *smused, pow2[3] ={1,2,4};
  char s[100];

  mn = max(len1,len2);
  matc = imatrix(3,mn+1);
  mused= imatrix(len1,len2);
  smused = musd;

  for (j=1;j<=len2;j++)
    for (i=1;i<=len1;i++)
      mused[i][j]=*musd++;
  SST=imatrix(len1, len2);
  for (j=0;j<=len2; j++)
    for (i=0;i<=len1;i++)
	SST[i][j]= *S10++;
   
 i = ii;
  j = jj;
  k = 1;
  while (i > 0 || j > 0)
    {      
      for (kk = 1; kk < 4; kk++)
	sets[kk] = SST[i][j] / pow2[kk-1] % 2;
      iiuse = 0;
      for (kk = 1; kk < 4; kk++) 
	if (sets[kk] == 1)
	  {
	    iiuse = kk;
	    break;
	  }
      if (iiuse == 1)
	{
	  reli = i - 1;
	  relj = j;
	  matc[1][k] = i + s1;
	  matc[2][k] = 0;
	}
      else if (iiuse == 2)
	{
	  reli = i-1;
	  relj = j-1;
	  matc[1][k] = i + s1;
	  matc[2][k] = j + s2;    
	  mused[i][j] = 1;	
	}
      else if (iiuse == 3) 
	{
	  reli = i;    
	  relj = j - 1;
	  matc[1][k] = 0;
	  matc[2][k] = j + s2;
	}
      i = reli;
      j = relj;
      endi = i;
      endj = j;
      if (iiuse == 0) 
	{
	  i = 0;  
	  j = 0;
	}
      else
	k = k + 1;
    }
  *mlen = k - 1;

  /* copy matc to mtc etc */
  for (j=1;j<k; j++)
    for (i=1;i<=2;i++)
      *mtc++ = matc[i][j];
  *eni=endi;
  *enj=endj;
  musd=smused;
  for (j=1;j<=len2;j++)
    for (i=1;i<=len1;i++)
      *musd++=mused[i][j];
  free_imatrix( matc,3,mn+1);
  free_imatrix(mused,len1,len2);
  free_imatrix(SST,len1,len2);  
}
/*
  sprintf(s,"mused %d %d ",reli,relj);
	  errmsg(s);
*/  
STDCALL void recompS (int li, int lj, int *seq1, int *seq2, int *musd,
			 int localal, int n, int m, int *S10,
			 double *S11, int *SN10, double *SN, double *SX, 
			 double *ss1, int sss, int *jm, 
		         int *im, double *maxs)
{
  double poss[3], tmp;
  double **SSSnew, **SSS, **ss, maxsim;
  int allposs, i, j, k, maxj, used, ii, **SSTnew, **SST,  **mused, iimax,jjmax;
  char s[100];

  SSSnew=matrix(n+1,m+1);
  SSS=matrix(n+1,m+1);
  SSTnew=imatrix(n+1,m+1);
  SST=imatrix(n+1,m+1);
  ss = matrix(sss+1,sss+1);
  for (j=1;j<=sss;j++)
    for (i=1;i<=sss;i++)
      ss[i][j] = *ss1++;	   
  mused=imatrix(n,m);
  for (j=1;j<=m;j++)
	  for(i=1;i<=n;i++)
		  mused[i][j]=*musd++;
  
  for (j=0;j<=m; j++)
    for (i=0;i<=n;i++)
      {
	SSSnew[i][j]=*S11;
	SSS[i][j]=*S11++;
	SSTnew[i][j]=*S10;
	SST[i][j]=*S10++;
      }
  maxj = -1;
  
  for (i = li;i<  n;i++)
    {
    for (j = lj;j<  m;j++)
      {	
	  if (j>maxj)
	      maxj=j;	  
	used = mused[i+1][j+1];
	if (used==1)
	  { 
	    poss[0] = SSSnew[i] [j + 1] - SX[seq1[i]-1];
	    poss[1] = 0;
	    poss[2] = SSSnew[i + 1][ j] - SX[seq2[j]-1];
	  }
	else 
	  {
	    poss[0] = SSSnew[i][j + 1] - SX[seq1[i]-1];
	    poss[1] = SSSnew[i][j] + ss[seq1[i]][seq2[j]];
	    poss[2] = SSSnew[i + 1][j] - SX[seq2[j]-1];
	  }
	tmp = poss[0];
	for (k = 1;k<3;k++) 
	  if ((poss[k] > tmp) && (k != 1 || used!=1))
	    tmp = poss[k];
	allposs = 0;
	for (k = 0;k<3;k++)
	  if (k != 1 || used!=1) 
	    if (fabs(poss[k] - tmp) < 0.0001)
	      allposs = allposs + pow(2.0,(double) k);
	if (localal==1 && tmp < 0.0001) 
	  {
	    SSSnew[i + 1][j + 1] = 0;
	    SSTnew[i + 1][j + 1] = 0;
	  }
	else
	  {
	    SSSnew[i + 1][j + 1] = tmp;
	    SSTnew[i + 1][j + 1] = allposs;
	  }
	  
	if (SSSnew[i + 1][j + 1] == SSS[i + 1][j + 1] && j >= maxj) 
	  if (j > maxj)
	    {
	      maxj = j;
	      break;
	    }
      }
    }
  /*find appropriate maximum */
  if (localal==1)
  {
      maxsim=0.0;
      iimax=0;
      jjmax=0;
      for (i=0;i<=n;i++)
	  for (j=0;j<=m;j++)
	      if (maxsim<SSSnew[i][j])
	      {
		  iimax=i;
		  jjmax=j;
		  maxsim= SSSnew[i][j];
	      }
  }
  else
  {
      jjmax=0;
      iimax=n;
      maxsim=SSSnew[n][0];
      for (j=1;j<=m+1;j++)
	  if(maxsim<SSSnew[n][j])
	  {
	      jjmax=j;
	      maxsim=SSSnew[i][j];
	  }
  }
  
  /* copy things to send back*/
  *jm=jjmax;
  *im=iimax;
  *maxs=maxsim;
  for (j=0;j<=m; j++)
    for (i=0;i<=n;i++)
      {
	*SN10++ = SSTnew[i][j];
	*SN++ = SSSnew[i][j];
      }
  free_matrix(SSSnew,n,m);
  free_imatrix(SSTnew,n,m);
  free_matrix(SSS,n,m);
  free_imatrix(SST,n,m);
  free_matrix(ss,sss+1,sss+1);
  free_imatrix(mused,n,m);
}

static double **matrix(nr,nc)
int nr,nc;
{
        int i;
        double **m;
	
        m = (double **)calloc((nr+1),sizeof(double*));
	
        if (!m) errmsg("allocation failure 1 in matrix()",1);
	else
        for(i = 0;i <= nr;i++) {
                m[i] = (double *)calloc((nc+1),sizeof( double));
                if (!m[i]) errmsg("allocation failure 2 in matrix()",1);
        }
        return m;
}
static void free_matrix(m,nr,nc)
double **m;
int nr,nc;
{
        int i;
 
        for(i = nr;i >= 0;i--) free(m[i]);
        free(m);
}

static int **imatrix(nr,nc)
int nr,nc;
{
        int i;
        int **m;
 
        m = (int **)calloc((nr+1), sizeof(int*));
        if (!m) errmsg("allocation failure 1 in imatrix()",1);
 else
        for(i = 0;i <= nr;i++) {
                m[i] = (int *)calloc((nc+1), sizeof(int));
                if (!m[i]) errmsg("allocation failure 2 in imatrix()",1);
        }
        return m;
}
static void free_imatrix(m,nr,nc)
int **m;
int nr,nc;
{
        int i;
 
        for(i = nr;i >= 0;i--) free(m[i]);
        free(m);
}








